==> lhcsmapi-1.3.41.tar.gz: lhcsmapi-1.3.41/lhcsmapi/pyedsl/QueryBuilder.py <==
import numpy as np
import pandas as pd
import datetime

from lhcsmapi.Time import Time
from lhcsmapi.dbsignal.SignalUtilities import SignalUtilities
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.dbquery.DbQueryBuilder import DbQueryBuilderFactory


def validate_time_definition(kwargs):
    t_start = validate_key_in_kwargs('t_start', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64,
                                     np.int64, **kwargs) or 0
    t_end = validate_key_in_kwargs('t_end', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64,
                                   np.int64, **kwargs)
    duration = validate_key_in_kwargs('duration', tuple, list, **kwargs)
    return Time.get_query_period_in_unix_time(t_start, t_end, duration, '')


def validate_key_in_kwargs(key, *key_types, **kwargs):
    result = None
    if key in kwargs:
        result = kwargs[key]
        match_found = False
        for key_type in key_types:
            match_found = match_found or isinstance(result, key_type)
        if not match_found:
            raise TypeError("You entered a {} keyword, but the keyword needs to be a {}.".format(key, key_types))

    return result


class Query(object):

    def set_db(self, db):
        self.db = db

    def set_circuit_type(self, circuit_type):
        self.circuit_type = circuit_type

    def set_t_start(self, t_start):
        self.t_start = t_start

    def set_t_end(self, t_end):
        self.t_end = t_end

    def set_timestamp(self, timestamp):
        self.timestamp = timestamp

    def get_time_params(self):
        if hasattr(self, 'timestamp') and self.db == 'PM':
            return {'eventTime': self.timestamp}
        elif hasattr(self, 't_start') and self.db in ['NXCALS', 'CALS', 'PM']:
            return {'t_start': self.t_start, 't_end': self.t_end}
        else:
            raise ValueError('Wrong time definition!')


class QueryBuilder(object):
    """ Builder class to query databases (PM, CALS, NXCALS).
        The general syntax is
        QueryBuilder().{with_pm()/with_cals()/with_nxcals()}.with_circuit_type().{with_duration()/with_timestamp()}.with_metadata().{signal_query()/feature_query()}

        - EXAMPLE PM (event query)
        t_start = '2015-01-13 16:59:11+01:00'
        t_end = '2015-01-13 17:15:46+01:00'

        source_timestamp_df = QueryBuilder().with_pm().with_circuit_type('RB').with_duration(t_start=t_start, t_end=t_end)\
                        .with_metadata(circuit_name=circuit_name, system='PC').event_query().df

        print(source_timestamp_df)
        >>>              source            timestamp
        >>> 0  RPTE.UA23.RB.A12  1421164827600000000

        - EXAMPLE PM (signal query)
        timestamp_fgc = 1421164827600000000
        source_fgc = 'RPTE.UA23.RB.A12'

        i_meas_df = QueryBuilder().with_pm().with_circuit_type('RB').with_timestamp(timestamp_fgc)\
                .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query()\
                .synchronize_time(timestamp_fgc).convert_index_to_sec().dfs[0]

        print(i_meas_df.head())
        >>>          STATUS.I_MEAS
        >>> -359.98      -0.150693
        >>> -359.96      -0.152328
        >>> -359.94      -0.151348
        >>> -359.92      -0.150366
        >>> -359.90      -0.149385

        - Example NXCALS (assuming connection to NXCALS cluster)::
        t_start = '2015-01-13 16:59:11+01:00'
        t_end = '2015-01-13 17:15:46+01:00'

        i_meas_df = QueryBuilder().with_nxcals().with_circuit_type('RB').with_duration(t_start=t_start, t_end=t_end).\
                with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query(spark).synchronize_time().convert_index_to_sec().dfs[0]

        print(i_meas_df.head())
        >>>       I_MEAS
        >>> 0.0     0.07
        >>> 25.5    1.13
        >>> 26.0    1.72
        >>> 26.5    2.82
        >>> 27.0    4.32

        - EXAMPLE CALS

        import pytimber
        ldb = pytimber.LoggingDB()

        t_start = '2015-01-13 16:59:11+01:00'
        t_end = '2015-01-13 17:15:46+01:00'

        i_meas_df = QueryBuilder().with_cals().with_circuit_type('RB').with_duration(t_start=t_start, t_end=t_end).\
                        with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query(ldb).synchronize_time().convert_index_to_sec().dfs[0]

        print(i_meas_df.head())
        >>>      RPTE.UA23.RB.A12:I_MEAS
        >>> 0.0                     0.07
        >>> 0.0                     1.13
        >>> 0.0                     1.72
        >>> 0.0                     2.82
        >>> 0.0                     4.32

        """

    _msg = 'Set database name using with_pm(), with_cals(ldb), with_nxcals(spark) method.'

    def __init__(self):
        self.query = Query()

    def __getattr__(self, name):
        if name not in ['query', 'with_pm', 'with_cals', 'with_nxcals']:
            raise AttributeError('{}() is not supported. {}'.format(name, self._msg))

    def __repr__(self):
        return self._msg

    def with_pm(self):
        self.query.set_db('PM')
        self.query.dbconnector = None
        return QueryBuilderWithDb(self.query)

    def with_cals(self, dbconnector):
        self.query.set_db('CALS')
        self.query.dbconnector = dbconnector
        return QueryBuilderWithDb(self.query)

    def with_nxcals(self, dbconnector):
        self.query.set_db('NXCALS')
        self.query.dbconnector = dbconnector
        return QueryBuilderWithDb(self.query)

class QueryBuilderWithDb(object):
    _msg = 'Set time definition: for PM signal query, with_timestamp(),' \
           '\n for PM event query or CALS, NXCALS signal query, with_duration()'

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if name not in ['query', 'with_timestamp', 'with_duration']:
            raise AttributeError('{}() is not supported. {}'.format(name, self._msg))

    def __repr__(self):
        return 'Database name properly set to {}. {}'.format(self.query.db, self._msg)

    def with_timestamp(self, timestamp):
        if self.query.db == 'PM':
            self.query.set_timestamp(Time.to_unix_timestamp(timestamp))
            return QueryBuilderWithDbTimeDef(self.query)
        else:
            raise ValueError('Timestamp can be only set for PM database, got {}'.format(self.query.db))

    def with_duration(self, **duration):
        t_start, t_end = validate_time_definition(duration)
        t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, '')

        self.query.set_t_start(t_start)
        self.query.set_t_end(t_end)

        return QueryBuilderWithDbTimeDef(self.query)


class QueryBuilderWithDbTimeDef(object):
    _msg = 'Set generic query parameter using with_query_parameters() method, ' \
           'or a circuit signal using with_circuit_type() method.'

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if name not in ['query', 'with_circuit_type', 'with_query_parameters']:
            raise AttributeError('{}() is not supported. \n{}'.format(name, self._msg))

    def __repr__(self):
        if hasattr(self.query, 'timestamp'):
            return 'Timestamp set to timestamp={}. {}'.format(self.query.timestamp, self._msg)
        else:
            return 'Query duration set to t_start={}, t_end={}. {}' \
                .format(self.query.t_start, self.query.t_end, self._msg)

    def with_circuit_type(self, circuit_type):
        # Check if circuit_type is supported
        self.query.set_circuit_type(circuit_type.upper())
        return QueryBuilderWithDbTimeDefCircuitType(self.query)

    def with_query_parameters(self, **kwargs):
        db = self.query.db
        is_timestamp = hasattr(self.query, 'timestamp')
        is_t_start = hasattr(self.query, 't_start')

        self.query.db_query_builder = DbQueryBuilderFactory.create(db, is_timestamp, is_t_start)
        self.query.db_query_builder.create_query_parameters(kwargs)

        return QueryBuilderWithDbTimeDefMetadata(self.query)


class QueryBuilderWithDbTimeDefCircuitType(object):
    _msg = 'Set metadata using with_metadata() method, e.g., ' \
           'with_metadata(circuit_name=\'RB.A12\', system=\'PC\', signal=\'I_MEAS\').'

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if name not in ['query', 'with_metadata']:
            raise AttributeError('{}() not supported. \n{}'.format(name, self._msg))

    def __repr__(self):
        return 'Circuit_type set to {}. {}'.format(self.query.circuit_type, self.msg)

    def with_metadata(self, **kwargs):
        db = self.query.db
        is_timestamp = hasattr(self.query, 'timestamp')
        is_t_start = hasattr(self.query, 't_start')
        circuit_type = self.query.circuit_type

        self.query.db_query_builder = DbQueryBuilderFactory.create(db, is_timestamp, is_t_start, circuit_type)
        self.query.db_query_builder.create_query_metadata(kwargs)
        self.query.db_query_builder.update_query_metadata(kwargs)

        return QueryBuilderWithDbTimeDefMetadata(self.query)


class QueryBuilderWithDbTimeDefMetadata(object):

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if hasattr(self.query, 'timestamp') and name not in ['query', 'signal_query']:
            raise AttributeError('{}() not supported, for PM with timestamp you can call signal_query()'.format(name))
        elif hasattr(self.query, 't_start') and self.query.db == 'PM' and name not in ['query', 'event_query']:
            raise AttributeError('{}() not supported, for PM with duration you can call event_query()'.format(name))
        elif name not in ['query', 'signal_query']:
            raise AttributeError('{}() is not supported, you can call signal_query()'.format(name))

    def __repr__(self):
        if hasattr(self.query, 'timestamp') and self.query.db == 'PM':
            return 'Query signal using signal_query() method, i.e., signal_query().'
        elif hasattr(self.query, 't_start') and self.query.db != 'PM':
            return 'Query signal using signal_query() method, i.e., signal_query().'
        else:
            return 'Query events using event_query() method, i.e., signal_query().'

    def signal_query(self):
        time_params = self.query.get_time_params()
        circuit_dfs = self.query.db_query_builder.execute_signal_query(time_params, dbconnector=self.query.dbconnector)

        return SignalProcessing(circuit_dfs, self.query.db)

    def event_query(self):
        time_params = self.query.get_time_params()
        event_source_timestamp_df = self.query.db_query_builder.execute_event_query(time_params)

        return EventProcessing(event_source_timestamp_df)

    def feature_query(self, features, **kwargs):
        time_params = self.query.get_time_params()
        return FeatureProcessing(self.query.db_query_builder.execute_feature_query(time_params, features, **kwargs,
                                                                                   dbconnector=self.query.dbconnector))

class FeatureProcessing(object):
    def __init__(self, df):
        self.df = df

class EventProcessing(object):

    def __init__(self, df):
        self.df = df

    def filter_source(self, circuit_name, system):
        if system in ['QH', 'QDS']:
            if 'RQ' in circuit_name:
                self.df['is_circuit_ok'] = self.df.apply(lambda row: (SignalMetadata.get_sector_names_from_components(
                    'RQ', 'MAGNET', row['source'])[0] == circuit_name) |
                                                                     (SignalMetadata.get_sector_names_from_components(
                                                                         'RQ', 'MAGNET', row['source'])[
                                                                          1] == circuit_name), axis=1)
            else:
                self.df['is_circuit_ok'] = self.df.apply(lambda row: (
                        SignalMetadata.get_sector_names_from_components('RB', 'MAGNET', row['source']) == circuit_name),
                                                         axis=1)

            self.df = self.df[self.df['is_circuit_ok']]
            self.df.drop(columns=['is_circuit_ok'], inplace=True)
            self.df.reset_index(drop=True, inplace=True)
        return self

    def drop_duplicate_source(self):
        self.df = self.df.drop_duplicates(subset=['source']).sort_values(by=['timestamp']).reset_index(drop=True)
        return self

    def sort_values(self, by):
        self.df.sort_values(by=by, inplace=True)
        self.df.reset_index(inplace=True, drop=True)
        return self


class SignalProcessing(object):

    def __init__(self, dfs, db):
        self.dfs = dfs
        self.db = db

    def __getattr__(self, name):
        method_list = [func for func in dir(SignalProcessing) if
                       callable(getattr(SignalProcessing, func)) and not func.startswith("__")]
        if name not in method_list:
            raise AttributeError(name + '{}() is not supported, you can call {}'.format(name, method_list))

    def __repr__(self):
        method_list = [func for func in dir(SignalProcessing) if
                       callable(getattr(SignalProcessing, func)) and not func.startswith("__")]
        return 'Available methods are {}'.format(method_list)

    def synchronize_time(self, t0=None):
        self.dfs = [SignalUtilities.synchronize_df(df, t0) if not df.empty else df for df in self.dfs]
        return self

    def convert_index_to_sec(self):
        self.dfs = [SignalUtilities.convert_indices_to_sec(df) if not df.empty else df for df in self.dfs]
        return self

    def create_col_from_index(self, col):
        for df in self.dfs:
            df[col] = df.index
        return self

    def filter_median(self, window=3):
        self.dfs = [df.rolling(window=window, min_periods=1).median() if not df.empty else df for df in self.dfs]
        return self

    def remove_values_for_time_less_than(self, time_threshold):
        self.dfs = [df[df.index > time_threshold] if not df.empty else df for df in self.dfs]
        return self

    def remove_initial_offset(self, time_threshold):
        self.dfs = [df.subtract(df[df.index < time_threshold].mean()) if not df.empty else df for df in self.dfs]
        return self

==> lhcsmapi-1.3.104.tar.gz: lhcsmapi-1.3.104/lhcsmapi/pyedsl/QueryBuilder.py <==
import numpy as np
import pandas as pd
import datetime

from lhcsmapi.Time import Time
from lhcsmapi.dbsignal.SignalUtilities import SignalUtilities
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.dbquery.DbQueryBuilder import DbQueryBuilderFactory


def validate_time_definition(kwargs):
    t_start = validate_key_in_kwargs('t_start', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64,
                                     np.int64, **kwargs) or 0
    t_end = validate_key_in_kwargs('t_end', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64,
                                   np.int64, **kwargs)
    duration = validate_key_in_kwargs('duration', tuple, list, **kwargs)
    return Time.get_query_period_in_unix_time(t_start, t_end, duration, '')


def validate_key_in_kwargs(key, *key_types, **kwargs):
    result = None
    if key in kwargs:
        result = kwargs[key]
        match_found = False
        for key_type in key_types:
            match_found = match_found or isinstance(result, key_type)
        if not match_found:
            raise TypeError("You entered a {} keyword, but the keyword needs to be a {}.".format(key, key_types))

    return result


class Query(object):

    def set_db(self, db):
        self.db = db

    def set_circuit_type(self, circuit_type):
        self.circuit_type = circuit_type

    def set_t_start(self, t_start):
        self.t_start = t_start

    def set_t_end(self, t_end):
        self.t_end = t_end

    def set_timestamp(self, timestamp):
        self.timestamp = timestamp

    def get_time_params(self):
        if hasattr(self, 'timestamp') and self.db == 'PM':
            return {'eventTime': self.timestamp}
        elif hasattr(self, 't_start') and self.db in ['NXCALS', 'CALS', 'PM']:
            return {'t_start': self.t_start, 't_end': self.t_end}
        else:
            raise ValueError('Wrong time definition!')


class QueryBuilder(object):
    """ Builder class to query databases (PM, CALS, NXCALS).
        The general syntax is
        QueryBuilder().{with_pm()/with_cals()/with_nxcals()}.with_circuit_type().{with_duration()/with_timestamp()}.with_metadata().{signal_query()/feature_query()}

        - EXAMPLE PM (event query)
        t_start = '2015-01-13 16:59:11+01:00'
        t_end = '2015-01-13 17:15:46+01:00'

        source_timestamp_df = QueryBuilder().with_pm().with_circuit_type('RB').with_duration(t_start=t_start, t_end=t_end)\
                        .with_metadata(circuit_name=circuit_name, system='PC').event_query().df

        print(source_timestamp_df)
        >>>              source            timestamp
        >>> 0  RPTE.UA23.RB.A12  1421164827600000000

        - EXAMPLE PM (signal query)
        timestamp_fgc = 1421164827600000000
        source_fgc = 'RPTE.UA23.RB.A12'

        i_meas_df = QueryBuilder().with_pm().with_circuit_type('RB').with_timestamp(timestamp_fgc)\
                .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query()\
                .synchronize_time(timestamp_fgc).convert_index_to_sec().dfs[0]

        print(i_meas_df.head())
        >>>          STATUS.I_MEAS
        >>> -359.98      -0.150693
        >>> -359.96      -0.152328
        >>> -359.94      -0.151348
        >>> -359.92      -0.150366
        >>> -359.90      -0.149385

        - Example NXCALS (assuming connection to NXCALS cluster)::
        t_start = '2015-01-13 16:59:11+01:00'
        t_end = '2015-01-13 17:15:46+01:00'

        i_meas_df = QueryBuilder().with_nxcals().with_circuit_type('RB').with_duration(t_start=t_start, t_end=t_end).\
                with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query(spark).synchronize_time().convert_index_to_sec().dfs[0]

        print(i_meas_df.head())
        >>>       I_MEAS
        >>> 0.0     0.07
        >>> 25.5    1.13
        >>> 26.0    1.72
        >>> 26.5    2.82
        >>> 27.0    4.32

        - EXAMPLE CALS

        import pytimber
        ldb = pytimber.LoggingDB()

        t_start = '2015-01-13 16:59:11+01:00'
        t_end = '2015-01-13 17:15:46+01:00'

        i_meas_df = QueryBuilder().with_cals().with_circuit_type('RB').with_duration(t_start=t_start, t_end=t_end).\
                        with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query(ldb).synchronize_time().convert_index_to_sec().dfs[0]

        print(i_meas_df.head())
        >>>      RPTE.UA23.RB.A12:I_MEAS
        >>> 0.0                     0.07
        >>> 0.0                     1.13
        >>> 0.0                     1.72
        >>> 0.0                     2.82
        >>> 0.0                     4.32

        """

    _msg = 'Set database name using with_pm(), with_cals(ldb), with_nxcals(spark) method.'

    def __init__(self):
        self.query = Query()

    def __getattr__(self, name):
        if name not in ['query', 'with_pm', 'with_cals', 'with_nxcals']:
            raise AttributeError('{}() is not supported. {}'.format(name, self._msg))

    def __repr__(self):
        return self._msg

    def with_pm(self):
        self.query.set_db('PM')
        self.query.dbconnector = None
        return QueryBuilderWithDb(self.query)

    def with_cals(self, dbconnector):
        self.query.set_db('CALS')
        self.query.dbconnector = dbconnector
        return QueryBuilderWithDb(self.query)

    def with_nxcals(self, dbconnector):
        self.query.set_db('NXCALS')
        self.query.dbconnector = dbconnector
        return QueryBuilderWithDb(self.query)

class QueryBuilderWithDb(object):
    _msg = 'Set time definition: for PM signal query, with_timestamp(),' \
           '\n for PM event query or CALS, NXCALS signal query, with_duration()'

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if name not in ['query', 'with_timestamp', 'with_duration']:
            raise AttributeError('{}() is not supported. {}'.format(name, self._msg))

    def __repr__(self):
        return 'Database name properly set to {}. {}'.format(self.query.db, self._msg)

    def with_timestamp(self, timestamp):
        if self.query.db == 'PM':
            self.query.set_timestamp(Time.to_unix_timestamp(timestamp))
            return QueryBuilderWithDbTimeDef(self.query)
        else:
            raise ValueError('Timestamp can be only set for PM database, got {}'.format(self.query.db))

    def with_duration(self, **duration):
        t_start, t_end = validate_time_definition(duration)
        t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, '')

        self.query.set_t_start(t_start)
        self.query.set_t_end(t_end)

        return QueryBuilderWithDbTimeDef(self.query)


class QueryBuilderWithDbTimeDef(object):
    _msg = 'Set generic query parameter using with_query_parameters() method, ' \
           'or a circuit signal using with_circuit_type() method.'

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if name not in ['query', 'with_circuit_type', 'with_query_parameters']:
            raise AttributeError('{}() is not supported. \n{}'.format(name, self._msg))

    def __repr__(self):
        if hasattr(self.query, 'timestamp'):
            return 'Timestamp set to timestamp={}. {}'.format(self.query.timestamp, self._msg)
        else:
            return 'Query duration set to t_start={}, t_end={}. {}' \
                .format(self.query.t_start, self.query.t_end, self._msg)

    def with_circuit_type(self, circuit_type):
        # Check if circuit_type is supported
        self.query.set_circuit_type(circuit_type.upper())
        return QueryBuilderWithDbTimeDefCircuitType(self.query)

    def with_query_parameters(self, **kwargs):
        db = self.query.db
        is_timestamp = hasattr(self.query, 'timestamp')
        is_t_start = hasattr(self.query, 't_start')

        self.query.db_query_builder = DbQueryBuilderFactory.create(db, is_timestamp, is_t_start)
        self.query.db_query_builder.create_query_parameters(kwargs)

        return QueryBuilderWithDbTimeDefMetadata(self.query)


class QueryBuilderWithDbTimeDefCircuitType(object):
    _msg = 'Set metadata using with_metadata() method, e.g., ' \
           'with_metadata(circuit_name=\'RB.A12\', system=\'PC\', signal=\'I_MEAS\').'

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if name not in ['query', 'with_metadata']:
            raise AttributeError('{}() not supported. \n{}'.format(name, self._msg))

    def __repr__(self):
        return 'Circuit_type set to {}. {}'.format(self.query.circuit_type, self.msg)

    def with_metadata(self, **kwargs):
        db = self.query.db
        is_timestamp = hasattr(self.query, 'timestamp')
        is_t_start = hasattr(self.query, 't_start')
        circuit_type = self.query.circuit_type

        self.query.db_query_builder = DbQueryBuilderFactory.create(db, is_timestamp, is_t_start, circuit_type)
        self.query.db_query_builder.create_query_metadata(kwargs)
        self.query.db_query_builder.update_query_metadata(kwargs)

        return QueryBuilderWithDbTimeDefMetadata(self.query)


class QueryBuilderWithDbTimeDefMetadata(object):

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if hasattr(self.query, 'timestamp') and name not in ['query', 'signal_query']:
            raise AttributeError('{}() not supported, for PM with timestamp you can call signal_query()'.format(name))
        elif hasattr(self.query, 't_start') and self.query.db == 'PM' and name not in ['query', 'event_query']:
            raise AttributeError('{}() not supported, for PM with duration you can call event_query()'.format(name))
        elif name not in ['query', 'signal_query']:
            raise AttributeError('{}() is not supported, you can call signal_query()'.format(name))

    def __repr__(self):
        if hasattr(self.query, 'timestamp') and self.query.db == 'PM':
            return 'Query signal using signal_query() method, i.e., signal_query().'
        elif hasattr(self.query, 't_start') and self.query.db == 'PM':
            return 'Query signal using event_query() method, i.e., event_query().'
        else:
            return 'Query events using event_query() method, i.e., signal_query().'

    def signal_query(self):
        time_params = self.query.get_time_params()
        circuit_dfs = self.query.db_query_builder.execute_signal_query(time_params, dbconnector=self.query.dbconnector)

        return SignalProcessing(circuit_dfs, self.query.db)

    def event_query(self):
        time_params = self.query.get_time_params()
        event_source_timestamp_df = self.query.db_query_builder.execute_event_query(time_params)

        return EventProcessing(event_source_timestamp_df)

    def feature_query(self, features, **kwargs):
        time_params = self.query.get_time_params()
        return FeatureProcessing(self.query.db_query_builder.execute_feature_query(time_params, features, **kwargs,
                                                                                   dbconnector=self.query.dbconnector))

class FeatureProcessing(object):
    def __init__(self, df):
        self.df = df

class EventProcessing(object):

    def __init__(self, df):
        self.df = df

    def filter_source(self, circuit_name, system):
        if system in ['QH', 'QDS']:
            if 'RQ' in circuit_name:
                self.df['is_circuit_ok'] = self.df.apply(lambda row: (SignalMetadata.get_sector_names_from_components(
                    'RQ', 'MAGNET', row['source'])[0] == circuit_name) |
                                                                     (SignalMetadata.get_sector_names_from_components(
                                                                         'RQ', 'MAGNET', row['source'])[
                                                                          1] == circuit_name), axis=1)
            else:
                self.df['is_circuit_ok'] = self.df.apply(lambda row: (
                        SignalMetadata.get_sector_names_from_components('RB', 'MAGNET', row['source']) == circuit_name),
                                                         axis=1)

            self.df = self.df[self.df['is_circuit_ok']]
            self.df.drop(columns=['is_circuit_ok'], inplace=True)
            self.df.reset_index(drop=True, inplace=True)
        return self

    def drop_duplicate_source(self):
        self.df = self.df.drop_duplicates(subset=['source']).sort_values(by=['timestamp']).reset_index(drop=True)
        return self

    def sort_values(self, by):
        self.df.sort_values(by=by, inplace=True)
        self.df.reset_index(inplace=True, drop=True)
        return self


class SignalProcessing(object):

    def __init__(self, dfs, db):
        self.dfs = dfs
        self.db = db

    def __getattr__(self, name):
        method_list = [func for func in dir(SignalProcessing) if
                       callable(getattr(SignalProcessing, func)) and not func.startswith("__")]
        if name not in method_list:
            raise AttributeError(name + '{}() is not supported, you can call {}'.format(name, method_list))

    def __repr__(self):
        method_list = [func for func in dir(SignalProcessing) if
                       callable(getattr(SignalProcessing, func)) and not func.startswith("__")]
        return 'Available methods are {}'.format(method_list)

    def synchronize_time(self, t0=None):
        self.dfs = [SignalUtilities.synchronize_df(df, t0) if not df.empty else df for df in self.dfs]
        return self

    def convert_index_to_sec(self):
        self.dfs = [SignalUtilities.convert_indices_to_sec(df) if not df.empty else df for df in self.dfs]
        return self

    def create_col_from_index(self, col):
        for df in self.dfs:
            df[col] = df.index
        return self

    def filter_median(self, window=3):
        self.dfs = [df.rolling(window=window, min_periods=1).median() if not df.empty else df for df in self.dfs]
        return self

    def remove_values_for_time_less_than(self, time_threshold):
        self.dfs = [df[df.index > time_threshold] if not df.empty else df for df in self.dfs]
        return self

    def remove_initial_offset(self, time_threshold):
        self.dfs = [df.subtract(df[df.index < time_threshold].mean()) if not df.empty else df for df in self.dfs]
        return self
